/*
* RTP/RTCP H264/MPEG-4 AAC Encoder
* Copyright (c) 2008 Marius Gligor
*
* Author: Marius Gligor <marius.gligor@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/
//---------------------------------------------------------------------------
#include "AVCEncoder.h"
//---------------------------------------------------------------------------
static char * const psy_names[] = { "film", "animation", "grain", "stillimage", "psnr", "ssim", 0 };
//---------------------------------------------------------------------------
//#define LOG
#ifdef LOG
FILE *fp = 0;
// ctor
CAVCEncoder::CAVCEncoder() : CAvc()
{
  fp = fopen("log.txt", "wt");
  // init params
  memset(&params, 0, sizeof(params));
  enc = 0;
}
//---------------------------------------------------------------------------

// dtor
CAVCEncoder::~CAVCEncoder()
{
  fclose(fp);
}
#else
// ctor
CAVCEncoder::CAVCEncoder() : CAvc()
{
  // init params
  memset(&params, 0, sizeof(params));
  enc = 0;
}
//---------------------------------------------------------------------------

// dtor
CAVCEncoder::~CAVCEncoder()
{
}
#endif
//---------------------------------------------------------------------------

int32_t CAVCEncoder::Open(AVRTP_CONTEXT *cntx)
{
  this->cntx = cntx;
  // RTP
  rtp.Create(cntx->v_host, cntx->v_port, AVC_PAYLOAD_TYPE);
  rtp.SetClock(CLOCK90KHZ);
  rtp.SetBufferSize();
  rtp.SetTTL(cntx->ttl);
  // parameters
  h264_threading_init();
  h264_param_default_preset(&params, x264_preset_names[cntx->preset], tune());
  params.i_fps_num = cntx->frame_rate;
  params.i_fps_den = 1;
  params.i_width = cntx->width;
  params.i_height = cntx->height;
  params.i_keyint_max = cntx->key_frame_rate;
  params.i_keyint_min = 0;
  params.i_threads = 0;   // implicit
  params.i_csp = X264_CSP_I420;
  params.b_annexb = 0;
  params.b_aud = 1;
  params.i_slice_max_size = MTU - RTPHDRSIZE;
  //params.i_slice_max_mbs = 256;
  //params.i_slice_count = 0; //cntx->v_bit_rate;
  // try a loosless
  if(cntx->preset == 10)
  {
    params.rc.i_bitrate = cntx->v_bit_rate;
    params.rc.i_qp_constant = 0;
  }
  // turbo
  if(cntx->preset != 9)
  {
    h264_param_apply_fastfirstpass(&params);
  }
  h264_param_apply_profile(&params, x264_profile_names[cntx->profile]);
  // open encoder
  enc = h264_encoder_open(&params);
  if(!enc)
  {
    return -1;
  }
  h264_encoder_parameters(enc, &params);
  if(!params.b_repeat_headers)
  {
    EncodeHeaders();
  }
  // clean
  memset(&pic_inp, 0, sizeof(pic_inp));
  memset(&pic_out, 0, sizeof(pic_out));
  h264_picture_alloc(&pic_inp, params.i_csp, params.i_width, params.i_height);
  i420_size = I420_SIZE(cntx->width, cntx->height);
  return 0;
}
//---------------------------------------------------------------------------
char *CAVCEncoder::tune()
{
  char buffer[80] = "";
  if(psy_names[cntx->tune])
  {
    strcat(buffer, psy_names[cntx->tune]);
  }
  if(cntx->tune1 == 0)
  {
    strcat(buffer, strlen(buffer) == 0 ? "fastdecode" : ",fastdecode");
  }
  if(cntx->tune2 == 0)
  {
    strcat(buffer, strlen(buffer) == 0 ? "zerolatency" : ",zerolatency");
  }
  return strlen(buffer) == 0 ? NULL : buffer;
}

//---------------------------------------------------------------------------
int32_t CAVCEncoder::EncodeHeaders()
{
  // Write SPS/PPS/SEI
  int nnal = 0;
  x264_nal_t *headers = NULL;
  if(h264_encoder_headers(enc, &headers, &nnal) > 0)
  {
    return (nnal == 1) ? single(headers, 1) : aggregate(headers, nnal);
  }
  return 0;
}

//---------------------------------------------------------------------------
int32_t CAVCEncoder::Encode(uint8_t *rgb, int32_t rgb_size)
{
  int32_t nnal = 0;
  x264_nal_t *nals = NULL;
  // is open ?
  if(enc)
  {
    // fill image
    if(rgb_size == i420_size)
    {
      memcpy(pic_inp.img.plane[0], rgb, rgb_size);
    }
    else
    {
      RGB2I420P((RGBTRIPLE *)rgb, FLIP, params.i_width, params.i_height);
    }
    // encode
    int size = h264_encoder_encode(enc, &nals, &nnal, &pic_inp, &pic_out);
    if(size > 0)
    {
      for(int32_t i = 0, m = 1; i < nnal; i++)
      {
        x264_nal_t *nal = &nals[i];
        single(nal, m);
        m = 0;
      }
      // encode and send nals
      //return (nnal == 1) ? single(nals, 1) : aggregate(nals, nnal);
    }
  }
  return 0;
}
//---------------------------------------------------------------------------
#ifdef LOG
int32_t CAVCEncoder::stap_size(x264_nal_t *nals,  int32_t nnal)
{
  fprintf(fp, "n=%u\n", nnal);
  int32_t size = 0;
  for(int32_t i = 0; i < nnal; i++)
  {
    x264_nal_t *nal = &nals[i];
    size += nal->i_payload;
    fprintf(fp, "\ttype=%u idc=%u size=%u\n",
      nal->i_type, nal->i_ref_idc, nal->i_payload);
    fflush(fp);
  }
  fprintf(fp, "\ttotal=%u\n", size);
  return size;
}
#else
int32_t CAVCEncoder::stap_size(x264_nal_t *nals,  int32_t nnal)
{
  int32_t size = 0;
  for(int32_t i = 0; i < nnal; i++)
  {
    x264_nal_t *nal = &nals[i];
    size += nal->i_payload;
  }
  return size;
}
#endif
//---------------------------------------------------------------------------
int32_t CAVCEncoder::aggregate(x264_nal_t *nals,  int32_t nnal)
{
  // alloc buffer
  uint8_t *out = (uint8_t *)malloc(stap_size(nals, nnal));
  out[0] = 0;
  // STAP-A header
  NALU_HDR *hdr = (NALU_HDR *)out;
  hdr->type = STAP_A;
  uint8_t *tmp = &out[1];
  // packetization send nals
  for(int32_t i = 0; i < nnal; i++)
  {
    x264_nal_t *nal = &nals[i];
    int32_t nal_size = nal->i_payload - 2;
    if(hdr->nri < nal->i_ref_idc)
    {
      hdr->nri = nal->i_ref_idc;
    }
    memcpy(tmp, nal->p_payload + 2, nal_size);
    tmp += nal_size;
  }
  // send packet
  uint32_t packet_size = tmp - out;
  if(packet_size > 1)
  {
    rtp.SendPacket(out, packet_size, 1);
    cntx->video_bw = bwt.get(RTCPHDRSIZE + packet_size);
  }
  //  free buffer
  free(out);
  return 0;
}
//---------------------------------------------------------------------------
#ifdef LOG
int32_t CAVCEncoder::single(x264_nal_t *nal, int m)
{
  fprintf(fp, "n=%u\n", 1);
  fprintf(fp, "\ttype=%u idc=%u size=%u\n",
    nal->i_type, nal->i_ref_idc, nal->i_payload);
  fflush(fp);
  //
  int32_t nal_size = nal->i_payload - 4;
  rtp.SendPacket(nal->p_payload + 4, nal_size, m);
  cntx->video_bw = bwt.get(RTCPHDRSIZE + nal_size);
  return 0;
}
#else
int32_t CAVCEncoder::single(x264_nal_t *nal, int m)
{
  int32_t nal_size = nal->i_payload - 4;
  rtp.SendPacket(nal->p_payload + 4, nal_size, m);
  cntx->video_bw = bwt.get(RTCPHDRSIZE + nal_size);
  return 0;
}
#endif
//---------------------------------------------------------------------------

int32_t CAVCEncoder::fragment(x264_nal_t *nal)
{
  uint8_t out[MAX_PAYLOAD];
  memset(out, 0, sizeof(out));
  int32_t size = sizeof(out) - 2;
  uint8_t *tmp = nal->p_payload + 5;
  int32_t nal_size = nal->i_payload - 5;
  cntx->video_bw = bwt.get(nal_size);
  // first fragment
  NALU_HDR *hdr = (NALU_HDR *)&out[0];
  hdr->type = FU_A;
  hdr->nri = nal->i_ref_idc;
  hdr->f = 1;
  // FU HDR
  FU_HDR *fu_hdr = (FU_HDR *)&out[1];
  fu_hdr->type = nal->i_type;
  fu_hdr->s = 0;
  // copy nal fragment
  memcpy(out + 2, tmp, size);
  tmp += size;
  nal_size -= size;
  // send packet
  rtp.SendPacket(out, sizeof(out), 1);
  // next fragments
  size = sizeof(out) - 1;
  while(nal_size > size)
  {
    out[0] = 0;
    // FU HDR
    FU_HDR *fu_hdr = (FU_HDR *)&out[0];
    fu_hdr->type = nal->i_type;
    memcpy(out + 1, tmp, size);
    tmp += size;
    nal_size -= size;
    if(nal_size == 0)
    {
      fu_hdr->e = 1;
    }
    // send packet
    rtp.SendPacket(out, sizeof(out), 0);
  }
  // last fragment
  if(nal_size > 0)
  {
    out[0] = 0;
    // FU HDR
    FU_HDR *fu_hdr = (FU_HDR *)&out[0];
    fu_hdr->type = nal->i_type;
    fu_hdr->e = 1;
    memcpy(out + 1, tmp, nal_size);
    // send packet
    rtp.SendPacket(out, nal_size + 1, 0);
  }
  return 0;
}
//---------------------------------------------------------------------------

int32_t CAVCEncoder::Close()
{
  if(enc)
  {
    h264_picture_clean(&pic_inp);
    h264_encoder_close(enc);
    enc = 0;
  }

  rtp.Destroy();

  return 0;
}
//---------------------------------------------------------------------------

void CAVCEncoder::RGB2I420P(RGBTRIPLE *rgb, uint8_t flip,
    int32_t width, int32_t height)
{
  // get pointers to the data
  RGBTRIPLE *tmp = rgb;
  uint8_t *yline = pic_inp.img.plane[0];
  uint8_t *uline = pic_inp.img.plane[1];
  uint8_t *vline = pic_inp.img.plane[2];
  // convert
  for(int32_t y = 0;y < height; y++)
  {
    // flip image ?
    if(flip)
    {
      tmp = rgb + width * (height - 1 - y);
    }
    // encode I420
    for(int32_t x = 0;x < width;x++)
    {
      *yline = Y(tmp->rgbtRed, tmp->rgbtGreen, tmp->rgbtBlue);
      if(!((x | y) & 1))
      {
        *uline = U(tmp->rgbtBlue, *yline);
        *vline = V(tmp->rgbtRed, *yline);
        uline++;
        vline++;
      }
      yline++;
      tmp++;
    }
  }
}
//---------------------------------------------------------------------------

