/*
* RTP/RTCP H264/MPEG-4 AAC Encoder
* Copyright (c) 2008 Marius Gligor
*
* Author: Marius Gligor <marius.gligor@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "Capture.h"
#include "Network.h"
#include "Main.h"

//---------------------------------------------------------------------------
CCapture::CCapture() : CBaseCapture()
{
  // video
  pVideoCapture = NULL;
  pVideoSampleGrabber = NULL;
  pVideoRenderer = NULL;
  // audio
  pAudioCapture = NULL;
  pAudioSampleGrabber = NULL;
  pAudioRenderer = NULL;
}
//---------------------------------------------------------------------------
CCapture::~CCapture()
{
  // audio
  SAFE_RELEASE(pAudioCapture);
  SAFE_RELEASE(pAudioSampleGrabber);
  SAFE_RELEASE(pAudioRenderer);
  // video
  SAFE_RELEASE(pVideoCapture);
  SAFE_RELEASE(pVideoSampleGrabber);
  SAFE_RELEASE(pVideoRenderer);
}
//---------------------------------------------------------------------------
BOOL CCapture::InitGraph(AVRTP_CONTEXT &p_cntx)
{
  long EvCode;
  // save the context
  cntx = &p_cntx;
  // Create the Filter Graph Manager.
  hr = ::CoCreateInstance(CLSID_FilterGraph,
        NULL,
        CLSCTX_INPROC,
        IID_IGraphBuilder,
        (void **)&pGraph);
  if(hr != S_OK)
  {
    ShowError();
    return FALSE;
  }

  // Create the VideoCapture Graph Builder.
  hr = ::CoCreateInstance(CLSID_CaptureGraphBuilder2,
        NULL,
        CLSCTX_INPROC,
        IID_ICaptureGraphBuilder2,
        (void **)&pBuilder);
  if(hr != S_OK)
  {
    ShowError();
    return FALSE;
  }

  // Associate the graph with the builder.
  hr = pBuilder->SetFiltergraph(pGraph);
  if(hr != S_OK)
  {
    ShowError();
    return FALSE;
  }

  // try to build audio
  if(!AddAudioFilters())
  {
    ShowError();
    return FALSE;
  }

  // try to build video
  if(cntx->v && !AddVideoFilters())
  {
    ShowError();
    return FALSE;
  }

  if(cntx->preview)
  {
    ConfigAudioRenderer();
    ConfigVideoRenderer();
  }

  // get control interface
  hr = pGraph->QueryInterface(IID_IMediaControl, (void**)&pControl);
  if(hr != S_OK)
  {
    ShowError();
    return FALSE;
  }

  //Timeout();

  return TRUE;
}
//---------------------------------------------------------------------------
// Video
BOOL CCapture::AddVideoSource()
{
  BOOL bResult = AddSource(CLSID_VideoInputDeviceCategory,
      cntx->video_source,
      &pVideoCapture,
      L"VideoCapture");
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::AddVideoRenderer()
{
  // Create Renderer
  GUID guid = (cntx->preview) ? CLSID_VideoRenderer : CLSID_NullRenderer;
  BOOL bResult = AddRenderer(guid, &pVideoRenderer, L"VideoRenderer");
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::AddVideoGrabber()
{
  ISampleGrabber *pGrabber = NULL;
  // get interface
  if(!AddGrabber(&pVideoSampleGrabber,
    L"VideoSampleGrabber", &pGrabber))
  {
    return FALSE;
  }
  // configure
  AM_MEDIA_TYPE mt;
  ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
  mt.majortype = MEDIATYPE_Video;
  mt.subtype = MEDIASUBTYPE_RGB24;
  mt.formattype = FORMAT_VideoInfo;

  while(TRUE)
  {
    hr = pGrabber->SetMediaType(&mt);
    if(hr != S_OK)
    {
      break;
    }

    hr = pGrabber->SetOneShot(FALSE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetBufferSamples(TRUE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetCallback((ISampleGrabberCB *)cntx->v_call, 1L);
    break;
  }

  SAFE_RELEASE(pGrabber);
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::MakeNormal()
{
  // set format
  if(!SetVideoFormat())
  {
    return FALSE;
  }

  // nu e DV incearca normal
  hr = pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Video,
        pVideoCapture,
        pVideoSampleGrabber,
        pVideoRenderer);

  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::MakeDVAudio()
{
  IBaseFilter *pDVSplitter = NULL;

  hr = pGraph->FindFilterByName(L"DV Splitter", &pDVSplitter);
  if(hr == S_OK)
  {
    hr = pBuilder->RenderStream(NULL,
            &MEDIATYPE_Audio,
            pDVSplitter,
            pAudioSampleGrabber,
            pAudioRenderer);
    SAFE_RELEASE(pDVSplitter);
  }
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::MakeDVVideo()
{
  BOOL bResult = FALSE;
  // try video interleave
  hr = pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE,
      &MEDIATYPE_Interleaved,
      pVideoCapture,
      pVideoSampleGrabber,
      pVideoRenderer);
  if(hr != S_OK)
  {
    bResult = MakeNormal();
  }
  else
  {
    cntx->v = 1;
    if(!cntx->a)
    {
      cntx->a = MakeDVAudio();
    }
    if(!ConfigDV())
    {
      return FALSE;
    }
    if(cntx->v && !ConfigVideoGrabber())
    {
      return FALSE;
    }
    if(cntx->a && !ConfigAudioGrabber())
    {
      return FALSE;
    }
    bResult = TRUE;
  }
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::ConfigDV()
{
  int half_rate = (cntx->frame_rate <= 15) ? 1 : 0;
  if(!ChangeFrameRate(half_rate))
  {
    return FALSE;
  }
  int dv_res = (cntx->width == 720) ? 1 : 0;
  return ChangeVideoSize(dv_res);
}
//---------------------------------------------------------------------------
BOOL CCapture::AddVideoFilters()
{
  // add video source
  if(!AddVideoSource())
  {
    return FALSE;
  }
  // grabber
  if(!AddVideoGrabber())
  {
    return FALSE;
  }
  // renderer
  if(!AddVideoRenderer())
  {
    return FALSE;
  }

  return MakeDVVideo();
}
//---------------------------------------------------------------------------
BOOL CCapture::SetVideoFormat()
{
  BOOL bResult = FALSE;
  AM_MEDIA_TYPE *pAMT = NULL;
  IAMStreamConfig *pConfig = NULL;

  hr = pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Video,
        pVideoCapture,
        IID_IAMStreamConfig,
        (void **)&pConfig);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pConfig->GetFormat(&pAMT);
  if(hr == S_OK)
  {
    LONGLONG FrameRate = (ONESECOND100NANO / cntx->frame_rate);
    if(pAMT->formattype == FORMAT_VideoInfo)
    {
      VIDEOINFOHEADER *vih = (VIDEOINFOHEADER *)pAMT->pbFormat;
      vih->bmiHeader.biWidth = cntx->width;
      vih->bmiHeader.biHeight = cntx->height;
      if(cntx->change_rate)
      {
        vih->AvgTimePerFrame = FrameRate;
      }
      hr = pConfig->SetFormat(pAMT);
      bResult = (hr == S_OK);
    }
    else
    {
      if(pAMT->formattype == FORMAT_VideoInfo2)
      {
        VIDEOINFOHEADER2 *vih = (VIDEOINFOHEADER2 *)pAMT->pbFormat;
        vih->bmiHeader.biWidth = cntx->width;
        vih->bmiHeader.biHeight = cntx->height;
        if(cntx->change_rate)
        {
            vih->AvgTimePerFrame = FrameRate;
        }
        hr = pConfig->SetFormat(pAMT);
        bResult = (hr == S_OK);
      }
    }
  }

  DeleteMediaType(pAMT);
  SAFE_RELEASE(pConfig);
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::VideoProperty(HWND hWnd)
{
  BOOL bResult = FALSE;
  if(pVideoCapture)
  {
    bResult = FilterProperty(hWnd, &pVideoCapture);
  }
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::ConfigVideoRenderer()
{
  BOOL bResult = FALSE;
  IVideoWindow *pWindow = NULL;

  if(!pVideoRenderer)
  {
    return FALSE;
  }
  hr = pVideoRenderer->QueryInterface(IID_IVideoWindow, (void **)&pWindow);
  if(hr != S_OK)
  {
    return FALSE;
  }

  while(TRUE)
  {
    hr = pWindow->put_Owner((OAHWND)cntx->hWnd);
    if(hr != S_OK)
    {
      break;
    }
    hr = pWindow->put_WindowStyle(WS_CHILD|WS_CLIPSIBLINGS);
    if(hr != S_OK)
    {
      break;
    }
    hr = pWindow->SetWindowPosition(cntx->r.left, cntx->r.top,
        cntx->r.right, cntx->r.bottom);
    if(hr != S_OK)
    {
      break;
    }
    hr = pWindow->put_Visible(OATRUE);
    break;
  }

  SAFE_RELEASE(pWindow);
  bResult = (hr == S_OK);
  return bResult;
}
//---------------------------------------------------------------------------
// Audio
BOOL CCapture::AddAudioSource()
{
  BOOL bResult = AddSource(CLSID_AudioInputDeviceCategory,
    cntx->audio_source,
    &pAudioCapture,
    L"AudioCapture");
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::AddAudioRenderer()
{
  GUID guid = (cntx->preview) ? CLSID_DSoundRender : CLSID_NullRenderer;
  BOOL bResult = AddRenderer(guid, &pAudioRenderer, L"AudioRenderer");
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::AddAudioGrabber()
{
  ISampleGrabber *pGrabber = NULL;
  // get interface
  if(!AddGrabber(&pAudioSampleGrabber, L"AudioSampleGrabber", &pGrabber))
  {
    return FALSE;
  }
  // Set the media type.
  AM_MEDIA_TYPE mt;
  ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
  mt.majortype = MEDIATYPE_Audio;
  mt.subtype = MEDIASUBTYPE_PCM;
  mt.formattype = FORMAT_WaveFormatEx;

  while(TRUE)
  {
    hr = pGrabber->SetMediaType(&mt);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetOneShot(FALSE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetBufferSamples(TRUE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetCallback((ISampleGrabberCB *)cntx->a_call, 1L);
    break;
  }

  SAFE_RELEASE(pGrabber);
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::SetBufferSize(IBaseFilter *pCapture)
{
  IAMBufferNegotiation *pBuffer = NULL;

  hr = pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Audio,
        pCapture,
        IID_IAMBufferNegotiation,
        (void **)&pBuffer);
  if(hr != S_OK)
  {
    return FALSE;
  }

  ALLOCATOR_PROPERTIES AllocProp;
  AllocProp.cbAlign = -1;  // -1 means no preference.
  AllocProp.cbBuffer = cntx->frame_size;
  AllocProp.cbPrefix = -1;
  AllocProp.cBuffers = -1;
  hr = pBuffer->SuggestAllocatorProperties(&AllocProp);

  SAFE_RELEASE(pBuffer);
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::AddAudioFilters()
{
  // add grabber
  if(!AddAudioGrabber())
  {
    return FALSE;
  }
  // add renderer
  if(!AddAudioRenderer())
  {
    return FALSE;
  }
  // if no sound stop
  if(!cntx->a)
  {
    return TRUE;
  }

  if(!AddAudioSource())
  {
    return FALSE;
  }

  if(!SetAudioFormat(pAudioCapture))
  {
    return FALSE;
  }

  if(!SetBufferSize(pAudioCapture))
  {
    return FALSE;
  }

  hr = pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Audio,
        pAudioCapture,
        pAudioSampleGrabber,
        pAudioRenderer);

  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::SetAudioFormat(IBaseFilter *pCapture)
{
  BOOL bResult = FALSE;
  AM_MEDIA_TYPE *pAMT = NULL;
  IAMStreamConfig *pConfig = NULL;

  hr = pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Audio,
        pCapture,
        IID_IAMStreamConfig,
        (void **)&pConfig);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pConfig->GetFormat(&pAMT);
  if(hr == S_OK)
  {
    if(pAMT->formattype == FORMAT_WaveFormatEx)
    {
      memcpy(pAMT->pbFormat, cntx->wfx, pAMT->cbFormat);
      hr = pConfig->SetFormat(pAMT);
      bResult = (hr == S_OK);
    }
  }

  DeleteMediaType(pAMT);
  SAFE_RELEASE(pConfig);
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::AudioProperty(HWND hWnd)
{
  BOOL bResult = FALSE;
  if(pAudioCapture)
  {
    bResult = FilterProperty(hWnd, &pAudioCapture);
  }
  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::ConfigAudioRenderer()
{
  IBasicAudio *pAudio = NULL;
  // audio
  if(!pAudioRenderer)
  {
    return FALSE;
  }
  // config
  hr = pAudioRenderer->QueryInterface(IID_IBasicAudio, (LPVOID *)&pAudio);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pAudio->put_Volume(0);

  SAFE_RELEASE(pAudio);
  return TRUE;
}
//---------------------------------------------------------------------------
void CCapture::ShowError()
{
//    ::MessageBox(NULL, GetLastError(), L"Capture error", MB_OK|MB_ICONSTOP);
}
//---------------------------------------------------------------------------
BOOL CCapture::ChangeFrameRate(BOOL bHalfFrameRate)
{
  BOOL bResult = FALSE;
  IBaseFilter *pDVSplitter = NULL;
  IDVSplitter *pIDVSplitter = NULL;

  if(!pGraph)
  {
    return FALSE;
  }

  while(TRUE)
  {
    hr = pGraph->FindFilterByName(L"DV Splitter", &pDVSplitter);
    if(hr != S_OK)
    {
      break;
    }

    hr = pDVSplitter->QueryInterface(IID_IDVSplitter, (void **)&pIDVSplitter);
    if(hr != S_OK)
    {
      break;
    }

    hr = pIDVSplitter->DiscardAlternateVideoFrames(bHalfFrameRate);
    bResult = (hr == S_OK);
    break;
  }

  SAFE_RELEASE(pDVSplitter);
  SAFE_RELEASE(pIDVSplitter);

  return bResult;
}
//---------------------------------------------------------------------------
int CCapture::MakeVideoSize(int nSize)
{
  switch(nSize)
  {
    case 0:
      return DVRESOLUTION_FULL;
    case 1:
      return DVRESOLUTION_HALF;
    case 2:
      return DVRESOLUTION_QUARTER;
    default:
      return DVRESOLUTION_DC;
  }
}
//---------------------------------------------------------------------------
BOOL CCapture::ChangeVideoSize(int nSize)
{
  BOOL bResult = FALSE;
  IBaseFilter *pDVDecoder = NULL;
  IIPDVDec *pIPDVDec = NULL;

  if(!pGraph)
  {
    return FALSE;
  }

  while(TRUE)
  {
    hr = pGraph->FindFilterByName(L"DV Video Decoder", &pDVDecoder);
    if(hr != S_OK)
    {
      break;
    }

    hr = pDVDecoder->QueryInterface(IID_IIPDVDec, (void **)&pIPDVDec);
    if(hr != S_OK)
    {
      break;
    }

    hr = pIPDVDec->put_IPDisplay(MakeVideoSize(nSize));
    break;
  }

  SAFE_RELEASE(pDVDecoder);
  SAFE_RELEASE(pIPDVDec);
  bResult = (hr == S_OK);

  return bResult;
}
//---------------------------------------------------------------------------
BOOL CCapture::ConfigAudioGrabber()
{
  AM_MEDIA_TYPE mt;
  ISampleGrabber *pGrabber = NULL;

  if(!pAudioSampleGrabber)
  {
    return FALSE;
  }

  // get interface pointer
  hr = pAudioSampleGrabber->QueryInterface(IID_ISampleGrabber,
        (LPVOID *)&pGrabber);
  if(hr != S_OK)
  {
    return FALSE;
  }

  while(TRUE)
  {
    hr = pGrabber->GetConnectedMediaType(&mt);
    if(hr != S_OK)
    {
      break;
    }

    if(mt.formattype != FORMAT_WaveFormatEx)
    {
      break;
    }

    memcpy(cntx->wfx, mt.pbFormat, sizeof(WAVEFORMATEX));
    break;
  }

  FreeMediaType(mt);
  SAFE_RELEASE(pGrabber);
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CCapture::ConfigVideoGrabber()
{
  AM_MEDIA_TYPE mt;
  ISampleGrabber *pGrabber = NULL;
  VIDEOINFOHEADER *pInfoHeader;

  if(!pVideoSampleGrabber)
  {
    return FALSE;
  }

  // get interface pointer
  hr = pVideoSampleGrabber->QueryInterface(IID_ISampleGrabber,
        (LPVOID *)&pGrabber);
  if(hr != S_OK)
  {
    return FALSE;
  }

  while(TRUE)
  {
    hr = pGrabber->GetConnectedMediaType(&mt);
    if(hr != S_OK)
    {
      break;
    }

    if(mt.formattype != FORMAT_VideoInfo)
    {
      break;
    }

    // configure
    pInfoHeader = (VIDEOINFOHEADER *)mt.pbFormat;
    cntx->width = pInfoHeader->bmiHeader.biWidth;
    cntx->height = pInfoHeader->bmiHeader.biHeight;

    if(pInfoHeader->AvgTimePerFrame)
    {
        cntx->frame_rate = ONESECOND100NANO / pInfoHeader->AvgTimePerFrame;
    }

    break;
  }

  FreeMediaType(mt);
  SAFE_RELEASE(pGrabber);
  return (hr == S_OK);
}
//---------------------------------------------------------------------------
void CCapture::Timeout()
{
  REFERENCE_TIME beg = 0;
  REFERENCE_TIME end = 30 * 60 * ONESECOND100NANO; // 5 minute
  GUID pin = PIN_CATEGORY_CAPTURE;

  if(pBuilder)
  {
    // video
    if(pVideoCapture)
    {
      pBuilder->ControlStream(&pin, NULL,
          pVideoCapture, &beg, &end, 0, 0);
    }
    // audio
    if(pAudioCapture)
    {
      pBuilder->ControlStream(&pin, NULL,
          pAudioCapture, &beg, &end, 0, 0);
    }
  }
}
//---------------------------------------------------------------------------

