/*
* RTP/RTCP H264/MPEG-4 AAC Encoder
* Copyright (c) 2008 Marius Gligor
*
* Author: Marius Gligor <marius.gligor@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/
//---------------------------------------------------------------------------

#include "Network.h"
//---------------------------------------------------------------------------

// RTP
RTPNetwork::RTPNetwork() : RTPTime()
{
}
//---------------------------------------------------------------------------

RTPNetwork::~RTPNetwork()
{
}
//---------------------------------------------------------------------------

void RTPNetwork::Open(uint8_t *dest, uint32_t port)
{
  uint32_t yes = 1;
  // rtp
  rtp_sk = ::socket(AF_INET, SOCK_DGRAM, 0);
  ::setsockopt(rtp_sk, SOL_SOCKET, SO_REUSEADDR,
    (uint8_t *)&yes, sizeof(yes));
  ::ioctlsocket(rtp_sk, FIONBIO, (u_long*)&yes);
  // rtcp
  rtcp_sk = ::socket(AF_INET, SOCK_DGRAM, 0);
  ::setsockopt(rtcp_sk, SOL_SOCKET, SO_REUSEADDR,
    (uint8_t *)&yes, sizeof(yes));
  ::ioctlsocket(rtcp_sk, FIONBIO, (u_long*)&yes);

  switch(kind)
  {
    case IP_BROADCAST:
      // rtp
      ::setsockopt(rtp_sk, SOL_SOCKET, SO_BROADCAST,
        (uint8_t *)&yes, sizeof(yes));
      // rtcp
      ::setsockopt(rtcp_sk, SOL_SOCKET, SO_BROADCAST,
        (uint8_t *)&yes, sizeof(yes));
      break;
    case IP_MULTICAST:
      mreq.imr_multiaddr.s_addr = inet_addr(dest);
      mreq.imr_interface.s_addr = htonl(INADDR_ANY);
      // rtp
      ::setsockopt(rtp_sk, IPPROTO_IP,
        IP_ADD_MEMBERSHIP, (uint8_t *)&mreq, sizeof(mreq));
      // rtcp
      ::setsockopt(rtcp_sk, IPPROTO_IP,
        IP_ADD_MEMBERSHIP, (uint8_t *)&mreq, sizeof(mreq));
      break;
  }

  // rtp
  memset(&rtp, 0, sizeof(rtp));
  rtp.sin_family = AF_INET;
  rtp.sin_addr.s_addr = inet_addr(dest);
  rtp.sin_port = htons(port);
  // rtcp
  memset(&rtcp, 0, sizeof(rtcp));
  rtcp.sin_family = AF_INET;
  rtcp.sin_addr.s_addr = inet_addr(dest);
  rtcp.sin_port = htons(port + 1);
}
//---------------------------------------------------------------------------

void RTPNetwork::Close()
{
  // drop multicast
  if(kind == IP_MULTICAST)
  {
    // rtp
    ::setsockopt(rtp_sk, IPPROTO_IP, IP_DROP_MEMBERSHIP,
      (int8_t *)&mreq, sizeof(mreq));
    // rtcp
    ::setsockopt(rtcp_sk, IPPROTO_IP, IP_DROP_MEMBERSHIP,
      (int8_t *)&mreq, sizeof(mreq));
  }
  // rtp
  ::shutdown(rtp_sk, SD_BOTH);
  ::closesocket(rtp_sk);
  // rtcp
  ::shutdown(rtcp_sk, SD_BOTH);
  ::closesocket(rtcp_sk);
}
//---------------------------------------------------------------------------

void RTPNetwork::Create(uint8_t *dest, uint32_t port, uint8_t pt)
{
  kind = Check(dest);
  Open(dest, port);
  // rtp
  memset(&rtp_hdr, 0, sizeof(rtp_hdr));
  base_rtp_time = random(); // random
  rtp_hdr.time_stamp = base_rtp_time;
  rtp_hdr.ssrc = random();    // random
  rtp_hdr.b.sequence = random();
  rtp_hdr.b.pt = pt;
  rtp_hdr.b.v = 2;
  // rtcp
  memset(&rtcp_hdr, 0, sizeof(rtcp_hdr));
  rtcp_hdr.ssrc = rtp_hdr.ssrc;
  rtcp_hdr.b.v = 2;
  rtcp_sr.pack_count = 0;
  rtcp_sr.bytes_count = 0;
  base_ntp_time = ntp_time();
  last_sr_time = base_ntp_time;
  // prepare CNAME
  memset(cname, 0, sizeof(cname));
  cname[0] = CNAME;
  sprintf(&cname[2], "%s", "gmax@%rtpencoder");
  cname[1] = strlen(&cname[2]);
}
//---------------------------------------------------------------------------

void RTPNetwork::Destroy()
{
  // send bye
  SendRTCPBye();
  Close();
}
//---------------------------------------------------------------------------
// set TTL
void RTPNetwork::SetTTL(uint32_t TTL)
{
  // rtp
  ::setsockopt(rtp_sk, IPPROTO_IP,
    IP_MULTICAST_TTL, (int8_t *)&TTL, sizeof(uint32_t));
  // rtcp
  ::setsockopt(rtcp_sk, IPPROTO_IP,
    IP_MULTICAST_TTL, (int8_t *)&TTL, sizeof(uint32_t));
}
//---------------------------------------------------------------------------
// set buffer size
void RTPNetwork::SetBufferSize(uint32_t maxbuf)
{
  // rtp
  ::setsockopt(rtp_sk, SOL_SOCKET,
    SO_SNDBUF, (int8_t *)&maxbuf, sizeof(uint32_t));
  // rtcp
  ::setsockopt(rtcp_sk, SOL_SOCKET,
    SO_SNDBUF, (int8_t *)&maxbuf, sizeof(uint32_t));
}
//---------------------------------------------------------------------------

int32_t RTPNetwork::SendPacket(uint8_t *message, uint32_t size, uint8_t m)
{
  // RTP Header
  rtp_hdr.b.m = m;
  rtp_hdr.b.sequence++;
  //rtp_hdr.time_stamp += clock_rate;
  if(m)
  {
    rtp_hdr.time_stamp = rtp_time();
  }
  // prepare the buffer
  uint32_t packet_size = RTPHDRSIZE + size;
  uint8_t *buffer = (uint8_t *)malloc(packet_size);
  // copy header
  uint32_t *dst = (uint32_t *)buffer;
  uint32_t *src = (uint32_t *)&rtp_hdr;
  dst[0] = htonl(src[0]);
  dst[1] = htonl(src[1]);
  dst[2] = htonl(src[2]);
  // copy packet
  memmove(buffer + RTPHDRSIZE, message, size);
  // send
  result = ::sendto(rtp_sk, buffer, packet_size,
    0, (struct sockaddr *)&rtp, sizeof(rtp));
  free(buffer);
  // RTCP report
  SendRTCPSr(size);
  return result;
}
//---------------------------------------------------------------------------

int32_t RTPNetwork::Check(uint8_t *dest)
{
  uint32_t ip = inet_addr(dest);
  // invalid address
  if(ip == INADDR_NONE)
  {
    return IP_ERROR;
  }
  // multicast
  if((ip & IP_MULTICAST_MASK) == IP_MULTICAST_BYTE)
  {
    return IP_MULTICAST;
  }
  // broadcast / unicast
  return ((ip & IP_BROADCAST_MASK) == IP_BROADCAST_MASK) ? IP_BROADCAST : IP_UNICAST;
}
//---------------------------------------------------------------------------

// RTCP
int32_t RTPNetwork::SendRTCPBye()
{
  // RTCP Header
  rtcp_hdr.b.pt = RTCP_BYE;
  rtcp_hdr.b.sc = 1;
  rtcp_hdr.b.length = 1;
  // prepare the buffer
  uint32_t packet_size = RTCPHDRSIZE;
  uint8_t *buffer = (uint8_t *)malloc(packet_size);
  // copy header
  uint32_t *dst = (uint32_t *)buffer;
  uint32_t *src = (uint32_t *)&rtcp_hdr;
  dst[0] = htonl(src[0]);
  dst[1] = htonl(src[1]);
  // send the packet
  result = ::sendto(rtcp_sk, buffer, packet_size,
    0, (struct sockaddr *)&rtcp, sizeof(rtcp));
  // return
  free(buffer);
  return result;
}
//---------------------------------------------------------------------------

int32_t RTPNetwork::SendRTCPSr(uint32_t size)
{
  // prepare
  rtcp_sr.bytes_count += size;
  rtcp_sr.pack_count++;
  // rtcp ntp time
  uint64_t time = ntp_time();
  if(!sr(time))
  {
    return -1;
  }
  ntp_time(rtcp_sr.ntp_s, rtcp_sr.ntp_m, time);
  // rtcp time stamp
  rtcp_sr.timestamp = base_rtp_time;
  rtcp_sr.timestamp += rescale(time - base_ntp_time);
  // RTCP Header
  rtcp_hdr.b.pt = RTCP_SR;
  rtcp_hdr.b.length = 6;
  rtcp_hdr.b.sc = 0;
  // prepare the buffer
  uint32_t packet_size = RTCPHDRSIZE + sizeof(rtcp_sr);
  uint8_t *buffer = (uint8_t *)malloc(packet_size);
  // copy header
  uint32_t *dst = (uint32_t *)buffer;
  uint32_t *src = (uint32_t *)&rtcp_hdr;
  dst[0] = htonl(src[0]);
  dst[1] = htonl(src[1]);
  // and body
  src = (uint32_t *)&rtcp_sr;
  dst[2] = htonl(src[0]);
  dst[3] = htonl(src[1]);
  dst[4] = htonl(src[2]);
  dst[5] = htonl(src[3]);
  dst[6] = htonl(src[4]);
  // send packet
  result = ::sendto(rtcp_sk, buffer, packet_size,
    0, (struct sockaddr *)&rtcp, sizeof(rtcp));
  // return
  free(buffer);
  SendRTCPSdes();
  return result;
}
//---------------------------------------------------------------------------

int32_t RTPNetwork::SendRTCPSdes()
{
  // RTCP Header
  rtcp_hdr.b.pt = RTCP_SDES;
  rtcp_hdr.b.sc = 1;
  rtcp_hdr.b.length = 1 + sizeof(cname) / 4;
  // prepare the buffer
  uint32_t packet_size = RTCPHDRSIZE + sizeof(cname);
  uint8_t *buffer = (uint8_t *)malloc(packet_size);
  // copy header
  uint32_t *dst = (uint32_t *)buffer;
  uint32_t *src = (uint32_t *)&rtcp_hdr;
  dst[0] = htonl(src[0]);
  dst[1] = htonl(src[1]);
  // CNAME
  memcpy(buffer + RTCPHDRSIZE, cname, sizeof(cname));
  // send the packet
  result = ::sendto(rtcp_sk, buffer, packet_size,
    0, (struct sockaddr *)&rtcp, sizeof(rtcp));
  // return
  free(buffer);
  return result;
}
//---------------------------------------------------------------------------

